package PScm::Env;

use strict;
use warnings;
use base qw(PScm);

sub new {
    my ($class, %bindings) = @_;

    $class = ref($class) || $class;
    bless { bindings => {%bindings}, parent => 0 }, $class;
}

sub Extend {
    my ($self, $ra_symbols, $ra_values) = @_;

    return $self->ExtendUnevaluated($ra_symbols,
        [map { $_->Eval($self) } @$ra_values]);
}

sub ExtendUnevaluated {
    my ($self, $ra_symbols, $ra_values) = @_;

    my %newbindings;
    @newbindings{ map { $_->value } @$ra_symbols } = @$ra_values;
    my $newenv = $self->new(%newbindings);
    $newenv->{parent} = $self;
    return $newenv;
}

sub ExtendRecursively {
    my ($self, $ra_symbols, $ra_values) = @_;

    my $newenv = $self->ExtendUnevaluated($ra_symbols, $ra_values);
    $newenv->_eval_values();
    return $newenv;
}

sub _eval_values {
    my ($self) = @_;
    map {
        $self->{bindings}{$_} =
          $self->{bindings}{$_}->Eval($self)
      }
      keys %{ $self->{bindings} };
}

sub LookUp {
    my ($self, $symbol) = @_;

    if (defined(my $ref = $self->_lookup_ref($symbol))) {
        return $$ref;
    } else {
        die "no binding for @{[$symbol->value]}",
          " in @{[ref($self)]}\n";
    }
}

sub Assign {
    my ($self, $symbol, $value) = @_;

    if (defined(my $ref = $self->_lookup_ref($symbol))) {
        $$ref = $value;
    } else {
        die "no binding for @{[$symbol->value]}",
          " in @{[ref($self)]}\n";
    }
}

sub _lookup_ref {
    my ($self, $symbol) = @_;

    if (exists($self->{bindings}{ $symbol->value })) {
        return \$self->{bindings}{ $symbol->value };
    } elsif ($self->{parent}) {
        return $self->{parent}->_lookup_ref($symbol);
    } else {
        return undef;
    }
}

1;
