package PScm::SpecialForm;

use strict;
use warnings;
use base qw(PScm::Expr);

###############################
package PScm::SpecialForm::Let;

use base qw(PScm::SpecialForm);

sub Apply {
    my ($self, $form, $env) = @_;

    my ($ra_symbols, $ra_values, $body) = $self->UnPack($form, $env);

    return $body->Eval($env->Extend($ra_symbols, $ra_values));
}

sub UnPack {
    my ($self, $form, $env) = @_;

    my ($bindings, $body) = $form->value;
    my (@symbols, @values);

    foreach my $binding ($bindings->value) {
        my ($symbol, $value) = $binding->value;
        push @symbols, $symbol;
        push @values,  $value;
    }

    return (\@symbols, \@values, $body);
}

##################################
package PScm::SpecialForm::LetRec;

use base qw(PScm::SpecialForm::Let);

sub Apply {
    my ($self, $form, $env) = @_;

    my ($ra_symbols, $ra_values, $body) = $self->UnPack($form, $env);

    return $body->Eval(
        $env->ExtendRecursively($ra_symbols, $ra_values));
}

##############################
package PScm::SpecialForm::If;

use base qw(PScm::SpecialForm);

sub Apply {
    my ($self, $form, $env) = @_;

    my ($condition, $true_branch, $false_branch) = $form->value;

    if ($condition->Eval($env)->isTrue) {
        return $true_branch->Eval($env);
    } else {
        return $false_branch->Eval($env);
    }
}

##################################
package PScm::SpecialForm::Lambda;

use base qw(PScm::SpecialForm);
use PScm::Closure;

sub Apply {
    my ($self, $form, $env) = @_;

    my ($args, $body) = $form->value;
    return PScm::Closure::Function->new($args, $body, $env);
}

1;
